//---------------------------------------------------------------------------
//-------------------------------------------------------------------
//	File:		   snd.c
//	Comments:   by George Ziniewicz
//-------------------------------------------------------------------

#include "snd.h"
#include "utility.h"

//--------------------------------------------------------------------------------------

Snd::Snd() {
	caps=0;
	numDrivers=0;
	sndDriver=-1;
	sndOutput=-1;
	hasSound=0;
	hasReverb=0;
	hasHw=0;
	on=0;

	lastTime=0;
	lastChannel=-1;

	volume=1;
	pan=.5;
	reverb=1;

	type=FSOUND_2D;
	mind=200;
	maxd=1200;

	sample1=0;
	glist=0;
	material=0;
	worldSize=1000;// units
	worldScale=1./(50./worldSize);// 1 meter / (50 meter world/1000 units), how many units in a meter
};

Snd::~Snd() {};

//----------------------------------------------------------------------------------
//----------------------------------------------------------------------------------
//----------------------------------------------------------------------------------

int Snd::init(void) {
	int i;
	char s[80];

	if (FSOUND_GetVersion() < FMOD_VERSION) { // check version
		sprintf(s,"Incorrect/outdated library version, should be FMOD %.02f\n", FMOD_VERSION);
		log_message(s);
		return 1;
	}
	FSOUND_SetOutput(FSOUND_OUTPUT_DSOUND); // set output
	sndOutput=FSOUND_GetOutput();
	if (sndOutput!=FSOUND_OUTPUT_DSOUND) {
		log_message("Snd: cannot set output to DirectSound");
		return 2;
	}
	numDrivers=FSOUND_GetNumDrivers(); // check drivers
	if (numDrivers==0) {
		log_message("Snd: no sound drivers");
		return 3;
	}
	for (i=0; i<numDrivers; i++) { // check capabilities
		caps=0;
		FSOUND_GetDriverCaps(i, &caps);
		if (caps & FSOUND_CAPS_HARDWARE) {
			hasHw=1;
			sndDriver=i;
			hasReverb=1;
			type=FSOUND_HW3D;
			FSOUND_3D_Listener_SetDistanceFactor(worldScale);//how many units in a meter
			break;
		}
	}
	if (sndDriver==-1) {
		log_message("Snd: no sound driver w/ required capabilities");
		return 4;
	}
	FSOUND_SetDriver(sndDriver);
	FSOUND_SetMixer(FSOUND_MIXER_QUALITY_AUTODETECT);
	if (!FSOUND_Init(44100, 32, 0)) {
		log_message("Snd: bad initilization:");
		log_error();
		return 5;
	}
	//if (on) FSOUND_SetSFXMasterVolume(volume);
	inited=1; hasSound=1; on=1; 
	return 0;
}

//----------------------------------------------------------------------------------

void Snd::deinit(void) {
	free_sample(sample1);
	close_geom();
	FSOUND_Close();
	return;
}

//----------------------------------------------------------------------------------

int Snd::play(const char * s) { // play a wav file
	if (!on) return -1;
	if (hasReverb) {
		play(s,hw->camera->x,
			   hw->camera->y,
			   hw->camera->z);
		return 0;
	}
	sample1=FSOUND_Sample_Load(FSOUND_FREE,s,FSOUND_2D,0);
	if (sample1) return finish_2d(sample1);
	return -1;
}

//-------------------------------------------------------------------------

int Snd::play(FSOUND_SAMPLE *sample) { // play a sample
	if ((!on)||(!sample)) return -1;
	if (hasReverb) {
		play(sample,hw->camera->x,
			        hw->camera->y,
					hw->camera->z);
		return 0;
	}
	return finish_2d(sample);
}

//-------------------------------------------------------------------------

int Snd::finish_2d(FSOUND_SAMPLE *sample) {
	lastChannel=FSOUND_PlaySound(FSOUND_FREE, sample);
	if (lastChannel<0) return -1;
	set_pan(pan);
	set_volume(volume);
	lastTime=hw->time->fTime;
	return 0;
}

//-------------------------------------------------------------------------

void Snd::set_volume(float involume) {
	if (!on) return;
	 if (lastChannel>=0) FSOUND_SetVolume(lastChannel,involume*255);
	return;
}

//-------------------------------------------------------------------------

void Snd::set_pan(float inpan) {
	if (!on) return;
	if (lastChannel>=0) FSOUND_SetPan(lastChannel,inpan*255);
	return;
}

//-------------------------------------------------------------------------

void Snd::set_frequency(int infrequency) {
	if (!on) return;
	frequency=infrequency;
	if (lastChannel>=0) FSOUND_SetFrequency(lastChannel,frequency);
	return;
}

//-------------------------------------------------------------------------

void Snd::free_sample(FSOUND_SAMPLE *sample) {
	if (sample) FSOUND_Sample_Free(sample);
	return;
}

//-------------------------------------------------------------------------

/*
int Snd::load_wav(const char *s, FSOUND_SAMPLE *sample) {
	sample = FSOUND_Sample_Load(FSOUND_FREE, s, FSOUND_2D, 0);
	if (!sample) {
		//printf("Error: %s\n", FMOD_ErrorString(FSOUND_GetError()));
		return 1;
	}
	return 0;
}
*/

//-------------------------------------------------------------------------
//-------------------------------------------------------------------------
//----------------------------------------------------- 3d ----------------

int Snd::play(const char * s, float x, float y, float z) { // play a wav file in 3d
	if (!on) return -1;
	if (!hasReverb) {
		play(s);
		return 0;
	}
	sample1=FSOUND_Sample_Load(FSOUND_FREE,s,FSOUND_HW3D,0);
	if (!sample1) return -1;
	return finish_3d(sample1,x,y,z);
}

//-------------------------------------------------------------------------

int Snd::play(FSOUND_SAMPLE *sample, float x, float y, float z) { // play a sample in 3d
	if ((!on)||(!sample))  return -1;
	if (!hasReverb) {
		play(sample);
		return 0;
	}
	return finish_3d(sample,x,y,z);
}

//-------------------------------------------------------------------------

int Snd::finish_3d(FSOUND_SAMPLE *sample, float x, float y, float z) {
	float pos[]={0,0,0},vel[]={0,0,0};

	pos[0]=x;
	pos[1]=y;
	pos[2]=-z;
	FSOUND_Sample_SetMinMaxDistance(sample, mind, maxd);// set sample volume range
	lastChannel=FSOUND_PlaySound(FSOUND_FREE, sample);  // play
	if (lastChannel<0) return -1;// err?
	FSOUND_3D_SetAttributes(lastChannel,pos,vel);// set channel attr
	FSOUND_Reverb_SetMix(lastChannel, FSOUND_REVERBMIX_USEDISTANCE);//set channel reverb
	lastTime=hw->time->fTime;// track time
	return 0;	
}

//-------------------------------------------------------------------------

int Snd::open_geom(void) {
	if (!hasReverb) return 1;// no 3d?
	if (glist) FSOUND_Geometry_List_Free(glist);//if used, reset
	glist=FSOUND_Geometry_List_Create(TRUE); // new geom list
	if (!glist)	{ // err?
		log_message("Snd: cannot create geom list");
		return 1;
	}
	FSOUND_Geometry_List_Begin(glist); // begin list
		if (material) FSOUND_Geometry_Material_Free(material);// if used, reset
		material = FSOUND_Geometry_Material_Create();         // new material
		FSOUND_Geometry_Material_SetAttributes(material, 1, 1, 0, 0); // set material
		FSOUND_Geometry_Material_Set(material);// update material
		make_geom_polys();               // get poly list
		FSOUND_Geometry_List_End(glist); // end list

	FSOUND_Geometry_AddList(glist);
	return 0;
}

//-------------------------------------------------------------------------
// COORDINATE SYSTEM : X = right, Y = up, Z = forwards. (Left handed)
void Snd::make_geom_polys(void) {
	int i,j,k;
	float poly[5][4][3]={
			-1,-1,-1,// left wall
			-1,-1, 1,
			-1, 1, 1,
			-1, 1,-1,
			 1,-1, 1,// right
			 1,-1,-1,
			 1, 1,-1,
			 1, 1, 1,
			-1,-1, 1,// front
			 1,-1, 1,
			 1, 1, 1,
			-1, 1, 1,
			 1,-1,-1,// rear
			-1,-1,-1,
			-1, 1,-1,
			 1, 1,-1,
			-1, 1, 1,// top
			 1, 1, 1,
			 1, 1,-1,
			-1, 1,-1};
	float normal[5][3]={
			 1, 0, 0, // left wall
			-1, 0, 0, // right wall
			 0, 0,-1, // front wall
			 0, 0, 1, // back wall
			 0,-1, 0};// top

	float size=worldSize*.5;

	for         (i=0; i<5; i++) {
		for     (j=0; j<4; j++) {
			for (k=0; k<3; k++) poly[i][j][k]*=size;
		}
	}

	for (i=0; i<4; i++) {
		FSOUND_Geometry_AddPolygon(poly[i][0], poly[i][1],
								   poly[i][2], poly[i][3],
									NULL, FSOUND_GEOMETRY_NORMAL |
									FSOUND_GEOMETRY_REFLECTIVE, 0);
	}
	return;
}

//---------------------------------------------------------------------------

void Snd::update_listener(void) {
	float pos[]={hw->camera->x,
		         hw->camera->y,
				-hw->camera->z},vel[]={0,0,0};

	if (!hasReverb) return;// no 3d?

	FSOUND_3D_Listener_SetAttributes(pos, vel,
		 hw->camera->forwardx,
		 hw->camera->forwardy,
		-hw->camera->forwardz,
		 hw->camera->upx,
		 hw->camera->upy,
		-hw->camera->upz);
	FSOUND_3D_Update();
}

//---------------------------------------------------------------------------

	//FSOUND_Geometry_Material_SetAttributes(material, 1, 1, 0.1f, 0.1f);
	//FSOUND_Geometry_Material_SetAttributes(material,0,0, 0,0);

//-------------------------------------------------------------------------

void Snd::close_geom(void) {
	if (!hasReverb) return;// no 3d?
	if (glist)    FSOUND_Geometry_List_Free(glist);
	if (material) FSOUND_Geometry_Material_Free(material);
	return;
}
		
//---------------------------------------------------------------------------

void Snd::set_reverb(int envn) {
	int env[]={
		FSOUND_PRESET_GENERIC,
		FSOUND_PRESET_PADDEDCELL,
		FSOUND_PRESET_ROOM,
		FSOUND_PRESET_BATHROOM,
		FSOUND_PRESET_LIVINGROOM,
		FSOUND_PRESET_STONEROOM,
		FSOUND_PRESET_AUDITORIUM,
		FSOUND_PRESET_CONCERTHALL,
		FSOUND_PRESET_CAVE,
		FSOUND_PRESET_ARENA,
		FSOUND_PRESET_HANGAR,
		FSOUND_PRESET_CARPETEDHALLWAY,
		FSOUND_PRESET_HALLWAY,
		FSOUND_PRESET_STONECORRIDOR,
		FSOUND_PRESET_ALLEY,
		FSOUND_PRESET_FOREST,
		FSOUND_PRESET_CITY,
		FSOUND_PRESET_MOUNTAINS,
		FSOUND_PRESET_QUARRY,
		FSOUND_PRESET_PLAIN,
		FSOUND_PRESET_PARKINGLOT,
		FSOUND_PRESET_SEWERPIPE,
		FSOUND_PRESET_UNDERWATER,
		FSOUND_PRESET_DRUGGED,
		FSOUND_PRESET_DIZZY,
		FSOUND_PRESET_PSYCHOTIC};

	if (!hasReverb) return;// no 3d?
	FSOUND_Reverb_SetEnvironment(env[envn%26],//1,20,1);
		FSOUND_REVERB_IGNOREPARAM, FSOUND_REVERB_IGNOREPARAM, FSOUND_REVERB_IGNOREPARAM);
	return;
}

//int env,float vol,float decay,float damp
		/*
		FSOUND_ENVIRONMENT_GENERIC,
		FSOUND_ENVIRONMENT_PADDEDCELL,
		FSOUND_ENVIRONMENT_ROOM,
		FSOUND_ENVIRONMENT_BATHROOM,
		FSOUND_ENVIRONMENT_LIVINGROOM,
		FSOUND_ENVIRONMENT_STONEROOM,
		FSOUND_ENVIRONMENT_AUDITORIUM,
		FSOUND_ENVIRONMENT_CONCERTHALL,
		FSOUND_ENVIRONMENT_CAVE,
		FSOUND_ENVIRONMENT_ARENA,
		FSOUND_ENVIRONMENT_HANGAR,
		FSOUND_ENVIRONMENT_CARPETEDHALLWAY,
		FSOUND_ENVIRONMENT_HALLWAY,
		FSOUND_ENVIRONMENT_STONECORRIDOR,
		FSOUND_ENVIRONMENT_ALLEY,
		FSOUND_ENVIRONMENT_FOREST,
		FSOUND_ENVIRONMENT_CITY,
		FSOUND_ENVIRONMENT_MOUNTAINS,
		FSOUND_ENVIRONMENT_QUARRY,
		FSOUND_ENVIRONMENT_PLAIN,
		FSOUND_ENVIRONMENT_PARKINGLOT,
		FSOUND_ENVIRONMENT_SEWERPIPE,
		FSOUND_ENVIRONMENT_UNDERWATER,
		FSOUND_ENVIRONMENT_DRUGGED,
		FSOUND_ENVIRONMENT_DIZZY,
		FSOUND_ENVIRONMENT_PSYCHOTIC,
		FSOUND_ENVIRONMENT_COUNT*/
//-------------------------------------------------------------------------------

void Snd::set_loc(int channel, float x, float y, float z) {
		float pos[]={0,0,0},vel[]={0,0,0};

	if (channel<0) return;// err?
	pos[0]=x;
	pos[1]=y;
	pos[2]=-z;
	FSOUND_3D_SetAttributes(channel,pos,vel);
	return;
}

//-------------------------------------------------------------------------------

void Snd::log_error(void) {
	log_message(FMOD_ErrorString(FSOUND_GetError()));
}

//-------------------------------------------------------------------------------


